'''
Created on 19.02.2023

@author: wf
'''
from tests.basetest import Basetest
import xmltodict
from urllib.request import urlopen

from meta.uml import PlantUml
from sidif.sidif import DataInterchange, SiDIFParser

class TestPlantUml(Basetest):
    """
    test plantuml handling
    """
    def testIssue4(self):
        '''
        https://github.com/WolfgangFahl/py-sidif/issues/4
        convert dict tree to sidif
        '''
        uri="https://raw.githubusercontent.com/semstats/semstats.github.io/master/2019/ceur/workshop.xml"
        xml=urlopen(uri).read().decode()
        wsdict=xmltodict.parse(xml)
        if self.debug:
            print(wsdict)
        dif=DataInterchange.ofDict(wsdict,context="ceurws")
        sidifStr=dif.asSiDIF()
        if self.debug:
            print(sidifStr)
        self.assertTrue("workshop isA Topic" in sidifStr)
        self.assertTrue("homepage isA Property" in sidifStr)
        uml=PlantUml(title="CEUR-WS",copyRight="© Christoph Lange and contributors 2012–2020")
        uml.fromDIF(dif)
        if self.debug:
            print(uml)
        self.assertTrue('''workshop " 1" -- "editors *" editor''' in uml.uml)
  
    def getPresentation(self):
        '''
        get the SiDIF parse result for the presentation example
        '''
        self.baseUrl = "https://raw.githubusercontent.com/BITPlan/org.sidif.triplestore/master/src/test/resources/sidif"    
        url = f"{self.baseUrl}/presentation.sidif" 
        sp = SiDIFParser(debug=self.debug)
        parsed, error = sp.parseUrl(url, title="Presentation")
        self.assertTrue(error is None)
        #self.debug=True
        if self.debug:
            sp.printResult(parsed)
        dif=parsed[0]
        return dif
              
    def testIssue5(self):
        '''
        https://github.com/WolfgangFahl/py-sidif/issues/5
        convert sidif to plantuml #5
        '''
        dif=self.getPresentation()
        debug=self.debug
        #debug=True
        uml=PlantUml(title="Presentation",copyRight="© BITPlan GmbH 2015-2023")
        uml.fromDIF(dif)
        if debug:
            print (uml)
        self.assertTrue("package Presentation {" in uml.uml)
        self.assertTrue("class Icon {" in uml.uml)
        self.assertTrue("author:Text" in uml.uml)


# Auto generated from ceur-ws.yaml by pythongen.py version: 0.9.0
# Generation date: 2023-03-23T07:28:10
# Schema: CeurwsSchema
#
# id: CeurwsSchema
# description:
# license:

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.metamodelcore import Bool, Decimal, ElementIdentifier, NCName, NodeIdentifier, URI, URIorCURIE, XSDDate, XSDDateTime, XSDTime

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
CEURWSSCHEMA = CurieNamespace('CeurwsSchema', 'CeurwsSchema/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
SHEX = CurieNamespace('shex', 'http://www.w3.org/ns/shex#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = CEURWSSCHEMA


# Types
class String(str):
    """ A character string """
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "string"
    type_model_uri = CEURWSSCHEMA.String


class Integer(int):
    """ An integer """
    type_class_uri = XSD.integer
    type_class_curie = "xsd:integer"
    type_name = "integer"
    type_model_uri = CEURWSSCHEMA.Integer


class Boolean(Bool):
    """ A binary (true or false) value """
    type_class_uri = XSD.boolean
    type_class_curie = "xsd:boolean"
    type_name = "boolean"
    type_model_uri = CEURWSSCHEMA.Boolean


class Float(float):
    """ A real number that conforms to the xsd:float specification """
    type_class_uri = XSD.float
    type_class_curie = "xsd:float"
    type_name = "float"
    type_model_uri = CEURWSSCHEMA.Float


class Double(float):
    """ A real number that conforms to the xsd:double specification """
    type_class_uri = XSD.double
    type_class_curie = "xsd:double"
    type_name = "double"
    type_model_uri = CEURWSSCHEMA.Double


class Decimal(Decimal):
    """ A real number with arbitrary precision that conforms to the xsd:decimal specification """
    type_class_uri = XSD.decimal
    type_class_curie = "xsd:decimal"
    type_name = "decimal"
    type_model_uri = CEURWSSCHEMA.Decimal


class Time(XSDTime):
    """ A time object represents a (local) time of day, independent of any particular day """
    type_class_uri = XSD.dateTime
    type_class_curie = "xsd:dateTime"
    type_name = "time"
    type_model_uri = CEURWSSCHEMA.Time


class Date(XSDDate):
    """ a date (year, month and day) in an idealized calendar """
    type_class_uri = XSD.date
    type_class_curie = "xsd:date"
    type_name = "date"
    type_model_uri = CEURWSSCHEMA.Date


class Datetime(XSDDateTime):
    """ The combination of a date and time """
    type_class_uri = XSD.dateTime
    type_class_curie = "xsd:dateTime"
    type_name = "datetime"
    type_model_uri = CEURWSSCHEMA.Datetime


class DateOrDatetime(str):
    """ Either a date or a datetime """
    type_class_uri = LINKML.DateOrDatetime
    type_class_curie = "linkml:DateOrDatetime"
    type_name = "date_or_datetime"
    type_model_uri = CEURWSSCHEMA.DateOrDatetime


class Uriorcurie(URIorCURIE):
    """ a URI or a CURIE """
    type_class_uri = XSD.anyURI
    type_class_curie = "xsd:anyURI"
    type_name = "uriorcurie"
    type_model_uri = CEURWSSCHEMA.Uriorcurie


class Uri(URI):
    """ a complete URI """
    type_class_uri = XSD.anyURI
    type_class_curie = "xsd:anyURI"
    type_name = "uri"
    type_model_uri = CEURWSSCHEMA.Uri


class Ncname(NCName):
    """ Prefix part of CURIE """
    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "ncname"
    type_model_uri = CEURWSSCHEMA.Ncname


class Objectidentifier(ElementIdentifier):
    """ A URI or CURIE that represents an object in the model. """
    type_class_uri = SHEX.iri
    type_class_curie = "shex:iri"
    type_name = "objectidentifier"
    type_model_uri = CEURWSSCHEMA.Objectidentifier


class Nodeidentifier(NodeIdentifier):
    """ A URI, CURIE or BNODE that represents a node in a model. """
    type_class_uri = SHEX.nonLiteral
    type_class_curie = "shex:nonLiteral"
    type_name = "nodeidentifier"
    type_model_uri = CEURWSSCHEMA.Nodeidentifier


# Class references



@dataclass
class Volume(YAMLRoot):
    """
    A Volume is a collection of papers mostly documenting the results of an academic event
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = CEURWSSCHEMA.Volume
    class_class_curie: ClassVar[str] = "CeurwsSchema:Volume"
    class_name: ClassVar[str] = "Volume"
    class_model_uri: ClassVar[URIRef] = CEURWSSCHEMA.Volume

    number: Optional[float] = None
    acronym: Optional[str] = None
    wikidataid: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    url: Optional[str] = None
    date: Optional[Union[str, XSDDate]] = None
    dblp: Optional[str] = None
    k10plus: Optional[str] = None
    urn: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.number is not None and not isinstance(self.number, float):
            self.number = float(self.number)

        if self.acronym is not None and not isinstance(self.acronym, str):
            self.acronym = str(self.acronym)

        if self.wikidataid is not None and not isinstance(self.wikidataid, str):
            self.wikidataid = str(self.wikidataid)

        if self.title is not None and not isinstance(self.title, str):
            self.title = str(self.title)

        if self.description is not None and not isinstance(self.description, str):
            self.description = str(self.description)

        if self.url is not None and not isinstance(self.url, str):
            self.url = str(self.url)

        if self.date is not None and not isinstance(self.date, XSDDate):
            self.date = XSDDate(self.date)

        if self.dblp is not None and not isinstance(self.dblp, str):
            self.dblp = str(self.dblp)

        if self.k10plus is not None and not isinstance(self.k10plus, str):
            self.k10plus = str(self.k10plus)

        if self.urn is not None and not isinstance(self.urn, str):
            self.urn = str(self.urn)

        super().__post_init__(**kwargs)


@dataclass
class Session(YAMLRoot):
    """
    A Session is a a collection of papers as part of a Volume
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = CEURWSSCHEMA.Session
    class_class_curie: ClassVar[str] = "CeurwsSchema:Session"
    class_name: ClassVar[str] = "Session"
    class_model_uri: ClassVar[URIRef] = CEURWSSCHEMA.Session

    title: Optional[str] = None
    volume: Optional[Union[dict, Volume]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.title is not None and not isinstance(self.title, str):
            self.title = str(self.title)

        if self.volume is not None and not isinstance(self.volume, Volume):
            self.volume = Volume(**as_dict(self.volume))

        super().__post_init__(**kwargs)


@dataclass
class Paper(YAMLRoot):
    """
    A paper is e.g. a scholarly article
    """
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = CEURWSSCHEMA.Paper
    class_class_curie: ClassVar[str] = "CeurwsSchema:Paper"
    class_name: ClassVar[str] = "Paper"
    class_model_uri: ClassVar[URIRef] = CEURWSSCHEMA.Paper

    description: Optional[str] = None
    id: Optional[str] = None
    wikidataid: Optional[str] = None
    title: Optional[str] = None
    pdfUrl: Optional[Union[str, URI]] = None
    volume: Optional[Union[dict, Volume]] = None
    session: Optional[Union[dict, Session]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.description is not None and not isinstance(self.description, str):
            self.description = str(self.description)

        if self.id is not None and not isinstance(self.id, str):
            self.id = str(self.id)

        if self.wikidataid is not None and not isinstance(self.wikidataid, str):
            self.wikidataid = str(self.wikidataid)

        if self.title is not None and not isinstance(self.title, str):
            self.title = str(self.title)

        if self.pdfUrl is not None and not isinstance(self.pdfUrl, URI):
            self.pdfUrl = URI(self.pdfUrl)

        if self.volume is not None and not isinstance(self.volume, Volume):
            self.volume = Volume(**as_dict(self.volume))

        if self.session is not None and not isinstance(self.session, Session):
            self.session = Session(**as_dict(self.session))

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass

slots.number = Slot(uri=CEURWSSCHEMA.number, name="number", curie=CEURWSSCHEMA.curie('number'),
                   model_uri=CEURWSSCHEMA.number, domain=None, range=Optional[float])

slots.acronym = Slot(uri=CEURWSSCHEMA.acronym, name="acronym", curie=CEURWSSCHEMA.curie('acronym'),
                   model_uri=CEURWSSCHEMA.acronym, domain=None, range=Optional[str])

slots.wikidataid = Slot(uri=CEURWSSCHEMA.wikidataid, name="wikidataid", curie=CEURWSSCHEMA.curie('wikidataid'),
                   model_uri=CEURWSSCHEMA.wikidataid, domain=None, range=Optional[str])

slots.title = Slot(uri=CEURWSSCHEMA.title, name="title", curie=CEURWSSCHEMA.curie('title'),
                   model_uri=CEURWSSCHEMA.title, domain=None, range=Optional[str])

slots.description = Slot(uri=CEURWSSCHEMA.description, name="description", curie=CEURWSSCHEMA.curie('description'),
                   model_uri=CEURWSSCHEMA.description, domain=None, range=Optional[str])

slots.url = Slot(uri=CEURWSSCHEMA.url, name="url", curie=CEURWSSCHEMA.curie('url'),
                   model_uri=CEURWSSCHEMA.url, domain=None, range=Optional[str])

slots.date = Slot(uri=CEURWSSCHEMA.date, name="date", curie=CEURWSSCHEMA.curie('date'),
                   model_uri=CEURWSSCHEMA.date, domain=None, range=Optional[Union[str, XSDDate]])

slots.dblp = Slot(uri=CEURWSSCHEMA.dblp, name="dblp", curie=CEURWSSCHEMA.curie('dblp'),
                   model_uri=CEURWSSCHEMA.dblp, domain=None, range=Optional[str])

slots.k10plus = Slot(uri=CEURWSSCHEMA.k10plus, name="k10plus", curie=CEURWSSCHEMA.curie('k10plus'),
                   model_uri=CEURWSSCHEMA.k10plus, domain=None, range=Optional[str])

slots.urn = Slot(uri=CEURWSSCHEMA.urn, name="urn", curie=CEURWSSCHEMA.curie('urn'),
                   model_uri=CEURWSSCHEMA.urn, domain=None, range=Optional[str])

slots.volume = Slot(uri=CEURWSSCHEMA.volume, name="volume", curie=CEURWSSCHEMA.curie('volume'),
                   model_uri=CEURWSSCHEMA.volume, domain=None, range=Optional[Union[dict, Volume]])

slots.id = Slot(uri=CEURWSSCHEMA.id, name="id", curie=CEURWSSCHEMA.curie('id'),
                   model_uri=CEURWSSCHEMA.id, domain=None, range=Optional[str])

slots.pdfUrl = Slot(uri=CEURWSSCHEMA.pdfUrl, name="pdfUrl", curie=CEURWSSCHEMA.curie('pdfUrl'),
                   model_uri=CEURWSSCHEMA.pdfUrl, domain=None, range=Optional[Union[str, URI]])

slots.session = Slot(uri=CEURWSSCHEMA.session, name="session", curie=CEURWSSCHEMA.curie('session'),
                   model_uri=CEURWSSCHEMA.session, domain=None, range=Optional[Union[dict, Session]])

slots.volume__number = Slot(uri=CEURWSSCHEMA.number, name="volume__number", curie=CEURWSSCHEMA.curie('number'),
                   model_uri=CEURWSSCHEMA.volume__number, domain=None, range=Optional[float])

slots.volume__acronym = Slot(uri=CEURWSSCHEMA.acronym, name="volume__acronym", curie=CEURWSSCHEMA.curie('acronym'),
                   model_uri=CEURWSSCHEMA.volume__acronym, domain=None, range=Optional[str])

slots.volume__wikidataid = Slot(uri=CEURWSSCHEMA.wikidataid, name="volume__wikidataid", curie=CEURWSSCHEMA.curie('wikidataid'),
                   model_uri=CEURWSSCHEMA.volume__wikidataid, domain=None, range=Optional[str])

slots.volume__title = Slot(uri=CEURWSSCHEMA.title, name="volume__title", curie=CEURWSSCHEMA.curie('title'),
                   model_uri=CEURWSSCHEMA.volume__title, domain=None, range=Optional[str])

slots.volume__description = Slot(uri=CEURWSSCHEMA.description, name="volume__description", curie=CEURWSSCHEMA.curie('description'),
                   model_uri=CEURWSSCHEMA.volume__description, domain=None, range=Optional[str])

slots.volume__url = Slot(uri=CEURWSSCHEMA.url, name="volume__url", curie=CEURWSSCHEMA.curie('url'),
                   model_uri=CEURWSSCHEMA.volume__url, domain=None, range=Optional[str])

slots.volume__date = Slot(uri=CEURWSSCHEMA.date, name="volume__date", curie=CEURWSSCHEMA.curie('date'),
                   model_uri=CEURWSSCHEMA.volume__date, domain=None, range=Optional[Union[str, XSDDate]])

slots.volume__dblp = Slot(uri=CEURWSSCHEMA.dblp, name="volume__dblp", curie=CEURWSSCHEMA.curie('dblp'),
                   model_uri=CEURWSSCHEMA.volume__dblp, domain=None, range=Optional[str])

slots.volume__k10plus = Slot(uri=CEURWSSCHEMA.k10plus, name="volume__k10plus", curie=CEURWSSCHEMA.curie('k10plus'),
                   model_uri=CEURWSSCHEMA.volume__k10plus, domain=None, range=Optional[str])

slots.volume__urn = Slot(uri=CEURWSSCHEMA.urn, name="volume__urn", curie=CEURWSSCHEMA.curie('urn'),
                   model_uri=CEURWSSCHEMA.volume__urn, domain=None, range=Optional[str])

slots.session__title = Slot(uri=CEURWSSCHEMA.title, name="session__title", curie=CEURWSSCHEMA.curie('title'),
                   model_uri=CEURWSSCHEMA.session__title, domain=None, range=Optional[str])

slots.session__volume = Slot(uri=CEURWSSCHEMA.volume, name="session__volume", curie=CEURWSSCHEMA.curie('volume'),
                   model_uri=CEURWSSCHEMA.session__volume, domain=None, range=Optional[Union[dict, Volume]])

slots.paper__description = Slot(uri=CEURWSSCHEMA.description, name="paper__description", curie=CEURWSSCHEMA.curie('description'),
                   model_uri=CEURWSSCHEMA.paper__description, domain=None, range=Optional[str])

slots.paper__id = Slot(uri=CEURWSSCHEMA.id, name="paper__id", curie=CEURWSSCHEMA.curie('id'),
                   model_uri=CEURWSSCHEMA.paper__id, domain=None, range=Optional[str])

slots.paper__wikidataid = Slot(uri=CEURWSSCHEMA.wikidataid, name="paper__wikidataid", curie=CEURWSSCHEMA.curie('wikidataid'),
                   model_uri=CEURWSSCHEMA.paper__wikidataid, domain=None, range=Optional[str])

slots.paper__title = Slot(uri=CEURWSSCHEMA.title, name="paper__title", curie=CEURWSSCHEMA.curie('title'),
                   model_uri=CEURWSSCHEMA.paper__title, domain=None, range=Optional[str])

slots.paper__pdfUrl = Slot(uri=CEURWSSCHEMA.pdfUrl, name="paper__pdfUrl", curie=CEURWSSCHEMA.curie('pdfUrl'),
                   model_uri=CEURWSSCHEMA.paper__pdfUrl, domain=None, range=Optional[Union[str, URI]])

slots.paper__volume = Slot(uri=CEURWSSCHEMA.volume, name="paper__volume", curie=CEURWSSCHEMA.curie('volume'),
                   model_uri=CEURWSSCHEMA.paper__volume, domain=None, range=Optional[Union[dict, Volume]])

slots.paper__session = Slot(uri=CEURWSSCHEMA.session, name="paper__session", curie=CEURWSSCHEMA.curie('session'),
                   model_uri=CEURWSSCHEMA.paper__session, domain=None, range=Optional[Union[dict, Session]])



'''
Created on 2022-11-23

@author: wf
'''
from datetime import datetime
from meta.mw import MediaWikiContext
from sidif.sidif import SiDIFParser
from lodstorage.jsonable import JSONAble
import sys
import typing
import urllib3

class MetaModelElement(JSONAble):
    '''
    a generic MetaModelElement
    
    to handle the technicalities of being a MetaModelElement so that derived
    MetaModelElements can focus on the MetaModel domain specific aspects
    '''
    
    def __init__(self):
        """
        construct me
        """
        self.__metamodel_props = {}
        cls = self.__class__
        if hasattr(cls, "getSamples"):
            for sample in cls.getSamples():
                for key in sample.keys():
                    if not key in self.__metamodel_props:
                        self.__metamodel_props[key] = key  # Property(self,key)
    
    def __str__(self):
        """
        get a string representation of me
        """
        text = self.__class__.__name__
        for prop_name in self.__metamodel_props.keys():
            if not isinstance(prop_name, str):
                pass
            elif hasattr(self, prop_name):
                value = getattr(self, prop_name)
                text += f"\n  {prop_name}={str(value)}"
        return text

    
class Context(MetaModelElement):
    """
    A Context groups some topics like a Namespace/Package.
    This class provides helper functions and constants to render a Context to corresponding wiki pages
    """       
    
    def __init__(self):
        """
        constructor
        """
        MetaModelElement.__init__(self)
        self.topics = {}
        self.errors = []
    
    @classmethod
    def getSamples(cls):
        samples = [{
            "name": "MetaModel",
            "since": datetime.strptime("2015-01-23", "%Y-%m-%d"),
            "copyright": "2015-2023 BITPlan GmbH",
            "master": "http://smw-meta.bitplan.com"
        }]
        return samples
    
    def error(self, msg):
        print(msg, file=sys.stderr)
        self.errors.append(msg)
        
    def lookupTopic(self,topic_name:str,purpose:str)->'Topic':
        """
        lookup the given topic_name in my topics for the given purpose
        
        Args:
            topic_name(str): the name of the topic to lookup
            purpose(str): the purpose to do the lookup for
        
        Returns:
            Topic: the topic if found else None and an error is added to my errors
        """
        if topic_name in self.topics:
            return self.topics[topic_name]
        else:
            self.error(f"""topic {topic_name} not found in context {getattr(self,"name","?")} for {purpose}""")
            return None
        
    def propertyAlreadyDeclared(self,prop_name:str,topic:'Topic',purpose:str)->bool:
        """
        check whether the given property or role name is already Declared
        
        Args:
            prop_name(str): the property to check for duplicates
            topic(Topic): the topic to check
            purpose(str): the purpose to be displayed in error messages
            
        Returns:
            bool: True if this prop_name has already been use
        """
        in_use=prop_name in topic.properties or prop_name in topic.targetTopicLinks
        if in_use:
            self.error(f"duplicate name {prop_name} in topic {topic.name} for {purpose}")
        return in_use
        
    def link_source_with_target(self,tl:'TopicLink',source:'Topic',target:'Topic'):
        """
        link the source with the target via the given topicLink
        
        Args:
            tl(TopicLink): the topicLink
            source(Topic): the source Topic
            target(Topic): the target Topic
        """
        ok=True
        ok=ok and not self.propertyAlreadyDeclared(tl.sourceRole,target,f"{tl.name}")
        ok=ok and not self.propertyAlreadyDeclared(tl.targetRole,source,f"{tl.name}")
        if ok:
            # n:m handling with two lists on each side
            source.sourceTopicLinks[tl.sourceRole]=tl
            source.targetTopicLinks[tl.targetRole]=tl
    

    def addLink(self,tl:'TopicLink'):
        """
        link source and target of the given topicLink
         
        Args:
            tl(TopicLink): the topicLink
            context(Context): the context in which the link "lives"
        """
        tl.sourceTopic=self.lookupTopic(tl.source,f"topicLink {tl.name}")
        tl.targetTopic=self.lookupTopic(tl.target,f"topicLink {tl.name}")
        if tl.targetTopic is not None and tl.sourceTopic is not None:
            self.link_source_with_target(tl, tl.sourceTopic, tl.targetTopic)   
            
    def addProperty(self,prop:'Property')->bool:
        """
        add the given property to this context
        
        Args:
            prop(Property): the property to add
        
        Returns:
            bool: True if the property adding was successful
        """
        if not hasattr(prop, "topic"):
            self.error(f"prop  {prop} has no topic")
            return False
        topic_name = prop.topic
        if not hasattr(prop, "name"):
            self.error(f"prop {prop} has no name")
            return False
        if not hasattr(prop, "type"):
            prop.type="Text"
        topic=self.lookupTopic(topic_name,f"property {prop.name}")
        if topic:
            topic.properties[prop.name] = prop  
            return True 
        
    def createProperty4TopicLink(self,tl:'TopicLink')->'Propertx':
        """
        create a property for the given topic link
        
        Args:
            tl: TopicLink - the topiclink to create a property for
        """
        # see https://wiki.bitplan.com/index.php/SiDIFTemplates#properties
        prop=Property()
        prop.name=tl.sourceRole
        prop.label=prop.name
        if hasattr(tl, "sourceDocumentation"):
            prop.documentation=tl.sourceDocumentation
        else:
            prop.documentation=f"{prop.name}"
        prop.topic=tl.target
        prop.type="Page"
        prop.topicLink=tl
        prop.isLink=True
        return prop
        
    @classmethod
    def fromDictOfDicts(cls, did:dict) -> 'Context':
        """
        fill me from the given dict of dicts
        
        Args:
            did(dict): the dict of dicts
            
        Returns:
            Context: the context read
        """
        context = None
        for _key, record in did.items():
            isA = record["isA"]
            if isA == "Context":
                context = Context()
                context.fromDict(record)
            elif isA == "Topic":
                topic = Topic()
                topic.fromDict(record)
                if context is None:
                    context = Context()
                    context.name = "GeneralContext"
                    context.since = "2022-11-26"
                    context.master = "http://master.bitplan.com"
                    context.error(f"topic {topic.name} has no defined context")
                if hasattr(topic, "name"):
                    context.topics[topic.name] = topic
                else:
                    context.error(f"missing name for topic {topic}")
            elif isA == "TopicLink":
                """
# Event n : 1 City
Event_in_City isA TopicLink
"eventInCity" is name of it
"city" is sourceRole of it
false is sourceMultiple of it
"City" is source of it
"event" is targetRole of it
true is targetMultiple of it
"Event" is target of it 
                """
                tl=TopicLink()
                tl.fromDict(record)
                context.addLink(tl)
            elif isA == "Property":
                prop = Property()
                prop.isLink=False
                prop.fromDict(record)
                context.addProperty(prop)
                    
        # link topic to concepts and add topicLinks
        for topic in context.topics.values():
            topic.setConceptProperty()
            for _tl_name,tl in topic.targetTopicLinks.items():
                prop=context.createProperty4TopicLink(tl)
                context.addProperty(prop)
                pass
        return context
    
    @classmethod
    def fromSiDIF_input(cls,sidif_input:str,debug:bool=False):
        """
        initialize me from the given SiDIF input which might be a file path
        or url
        
        Args:
            sidif_input: path to local file or URL
            debug(bool): if True swith debugging on
            
        Returns:
            Tuple[Context,str,str]: context, error and errorMessage
        """
        if sidif_input.startswith("http:") or sidif_input.startswith("https:"):
            url=sidif_input
            http = urllib3.PoolManager()
            response = http.request('GET', url)
            sidif = response.data.decode('utf-8')
        else:
            sidif_path=sidif_input
            with open(sidif_path, 'r') as sidif_file:
                sidif=sidif_file.read()
        return Context.fromSiDIF(sidif, title=sidif_input, debug=debug)
    
    @classmethod
    def fromSiDIF(cls,sidif:str,title:str,depth:int=None,debug:bool=False)->typing.Tuple['Context',object,str]:
        """
        initialize me from the given SiDIF markup
        
        Args:
            sidif(str): the SiDIF markup to parse
            title(str): the title for the SiDIF
            depth(int): the explain depth to show for the errorMessage
            debug(bool): if True handle debugging
            
        Returns:
            Tuple[Context,str,str]: context, error and errorMessage
            
        """
        errMsg = None
        context = None
        sp = SiDIFParser(debug=debug)
        parsed, error = sp.parseText(sidif, title,depth=depth)
        if debug:
            sp.printResult(parsed)
        if error is None:
            dif = parsed[0]
            did = dif.toDictOfDicts()
            context = Context.fromDictOfDicts(did)
            context.dif=dif
            context.did=did
        else:
            errMsg=sp.errorMessage(title,error, depth=depth)
        return context, error,errMsg
        
    @classmethod
    def fromWikiContext(cls, mw_context:MediaWikiContext, depth:int=None,debug:bool=False) -> 'MetaModel':
        """
        initialize me from the given MediaWiki Context
        
        Args:
            mw_context(MediaWikiContext): the Mediawiki context
            depth(int): the explain depth to show for the errorMessage
            debug(bool): if True handle debugging
            
        Return:
            tuple(MetaModel,Exception,str): the metamodel and potential parsing errors as Exception and error Message
        """
        context = None
        error = None
        errMsg = None
        sidif = None
        if debug:
            print(f"reading sidif for {mw_context.context} from {mw_context.wikiId}")
        try:
            sidif = mw_context.read_sidif()
        except BaseException as ex:
            error = ex
            errMsg=str(ex)
        if sidif is not None:
            context,error,errMsg=cls.fromSiDIF(sidif=sidif, title=mw_context.wikiId,depth=depth,debug=debug)
        return context, error,errMsg
    
class Topic(MetaModelElement):
    """
    A Topic is a Concept/Class/Thing/Entity
    """
    
    def __init__(self):
        """
        constructor
        """
        MetaModelElement.__init__(self)
        self.properties = {}
        self.sourceTopicLinks={}
        self.targetTopicLinks={}
    
    @classmethod
    def getSamples(cls):
        samples = [{
                "pageTitle": "Concept:Topic",
                "name": "Topic",
                "pluralName": "Topics",
                "icon": "File:Topic_icon.png",
                "iconUrl": "/images/a/ae/Index.png",
                "documentation": "A Topic is a Concept/Class/Thing/Entity",
                "wikiDocumentation": "A Topic is a Concept/Class/Thing/Entity",
                "context": "MetaModel",
                "listLimit": 7,
                "cargo": True
                }]
        return samples
    
    def setConceptProperty(self):
        """ 
        set my concept property to any primary key or mandatory property
        """
        self.conceptProperty=None 
        for prop in self.properties.values():
            mandatory=False
            primaryKey=False
            if hasattr(prop, "mandatory"):
                mandatory=prop.mandatory
            if hasattr(prop,"primaryKey"):
                primaryKey=prop.primaryKey
            if mandatory or primaryKey:
                self.conceptProperty=prop
                break
    
    def getPluralName(self) -> str:
        """
        get the plural name for this topic
                
        Returns:
            str: the pluralname e.g. "Topics" for "Topic" or "Status" for "Status" or
            "Entities" for "Entity"
        """
        plural_name = self.pluralName if hasattr(self, "pluralName") else f"{self.name}s"
        return plural_name
    
    def getListLimit(self) -> int:
        """
        get the list limit for this topic
        """
        listLimit = getattr(self, "listLimit", 200)
        return listLimit
    
    def sortProperties(self) -> list:
        """
        get the properties that we should sort by
        
        Returns:
            list: a list of properties in sort order
        """
        prop_list = []
        for prop in self.properties.values():
            if hasattr(prop, "sortPos"):
                sortPos = prop.sortPos
                if sortPos:
                    prop_list.append(prop)
        prop_list = sorted(prop_list, key=lambda prop: prop.sortPos)
        return prop_list
    
    def propertiesByIndex(self) ->list:
        """
        return the properties by index
        
        Returns:
            list: the list of properties sorted by index
        """
        def index(prop:'Property')->int:
            if hasattr(prop,"index"):
                return int(prop.index)
            else:
                return sys.maxsize
        prop_list=sorted(self.properties.values(),key=lambda prop:index(prop))
        return prop_list
    
    def askSort(self)->str:
        """
        generate the sort clause for my SMW ask query
        
        Returns:
            str: the generated wiki markup 
        """
        sort=""
        order=""
        delim=""
        sortproperties=self.sortProperties()
        for prop in sortproperties:
            direction="ascending" if getattr(prop,"sortAscending",True) else "descending"
            sort+=delim+f"{self.name} {prop.name}"
            order+=delim+direction
            delim=","
            pass
        sortClause=f"|sort={sort}\n" if sort else ""
        orderClause=f"|order={order}\n" if order else ""
        markup=f"{sortClause}{orderClause}"
        return markup
    
    def askQuery(self,mainlabel:str=None,filterShowInGrid:bool=True,listLimit:int=None)->str:
        """
        get the askQuery for the me topic
        
        Args:
            mainlabel(str): the mainlabel to use - topic.name as default
            filterShowInGrid(bool): if True include only properties with showInGrid not being false
            listLimit(int): the list limit to use
        Returns:
            str: the markup for the query
        """
        if listLimit is None:
            listLimit=self.getListLimit()
        if mainlabel is None:
            mainlabel=self.name
        markup=f"""{{{{#ask: [[Concept:{self.name}]]
|mainlabel={mainlabel}
"""
        for prop in self.properties.values():
            if filterShowInGrid and hasattr(prop, "showInGrid"):
                show=prop.showInGrid
            else:
                show=True
            if show:
                markup+=f"|?{self.name} {prop.name} = {prop.name}\n"
        markup+=f"|limit={listLimit}\n"
        markup+=f"""{self.askSort()}}}}}"""
        return markup

    
class Property(MetaModelElement):
    """
    Provides helper functions and constants for properties
    """
    
    def __init__(self):
        """
        constructor
        """
        MetaModelElement.__init__(self)

    @classmethod
    def getSamples(cls):
        samples = [
            {
                "name": "Title",
                "label": "Offical Name",
                "type": "Special:Types/Text",
                "index": 2,
                "sortPos": 2,
                "primaryKey": False,
                "mandatory": True,
                "namespace": "Test",
                "size": 25,
                "uploadable": False,
                "defaultValue": "Some Title",
                "inputType": "combobox",
                "allowedVaues": "Tile A, Title B",
                "documentation": " Documentation can contain\n line breaks",
                "values_from": "property=Title",
                "showInGrid": False,
                "isLink": False,
                "nullable": False,
                "topic": "Concept:Event"
            },
            {
                "name": "wikidataid",
                "type": "Special:Types/External identifier",
                "formatterURI":"https://www.wikidata.org/wiki/$1"
            },
            # Properties that are not included in the MetaModel
            {
                "placeholder": "e.g. SMWCon",
                "regexp": "NaturalNumber",
                "usedFor": "Concept:Event, Concept:Event series",
                "pageTitle": "Property:Title"
            }
        ]
        return samples

    
class TopicLink(MetaModelElement):
    """
    A TopicLink links two Concepts/Topics
    """

    @classmethod
    def getSamples(cls):
        samples = [
            {
                "isA": "TopicLink",
                "name": "containedProperties",
                "sourceRole": "topic",
                "sourceMultiple": False,
                "source": "Topic",
                "targetRole": "properties",
                "targetMultiple": True,
                "target": "Property"
            },
            {
                "isA": "TopicLink",
                "name": "containedTopics",
                "sourceRole": "context",
                "sourceMultiple": False,
                "source": "Context",
                "targetRole": "topics",
                "targetMultiple": True,
                "target": "Topic"
            },
            {
                "isA": "TopicLink",
                "name": "typeOfProperty",
                "sourceRole": "usedByProperties",
                "sourceMultiple": True,
                "source": "Property",
                "sourceDocumentation": "the properties having this type",
                "targetRole": "smw_type",
                "targetMultiple": False,
                "target": "SMW_Type",
                "masterDetail": False,
                "targetDocumentation": "the smw_type being used by this property"
            }    
        ]
        return samples



from dataclasses import dataclass
from typing import Optional
import dacite
@dataclass
class SMW_Type:
    """
    an SMW_Type is a data type which determines the possible values for that type e.g. a Boolean can hold true/false values while a Number can hold 3.1459 or 20. A Page can hold the name of a Wiki page see https://semantic-mediawiki.org/wiki/Help:List_of_datatypes
    """
    pageTitle:str
    type:Optional[str] # The Semantic MediaWiki type  without the prefix e.g. Text, Number, Boolean
    documentation:Optional[str] # The documentation of this Semantic Media Wiki type
    id:Optional[str] # SMW internal id of the type
    helppage:Optional[str] # The url of the 'official' documentation page of this type
    typepage:Optional[str] # The Semantic Media Wiki Special page for this specific type e.g. Special:Types/Text, Special:Types/Boolean, Special:Types/Date, Special:Types/Number, Special:Types/Page
    javaType:Optional[str] # Java mapping of this type

    @classmethod
    def askQuery(cls):
        """
        get the ask Query for SMW_Type
        
        Returns:
            str: the mediawiki markup for the ask query
        """
        ask="""{{#ask: [[Concept:SMW_Type]]
|mainlabel=pageTitle
|?SMW_Type type = type
|?SMW_Type documentation = documentation
|?SMW_Type id = id
|?SMW_Type helppage = helppage
|?SMW_Type typepage = typepage
|?SMW_Type javaType = javaType
| limit=200
}}"""
        return ask
        
    @classmethod
    def fromDict(cls,data:dict):
        """
        create a SMW_Type from the given dict
        
        Args:
            data(dict): the dict to create the SMW_Type from
        
        Returns:
            SMW_Type: the freshly created SMW_Type
        """
        smw_type=dacite.from_dict(data_class=cls,data=data)
        return smw_type


'''
Created on 2023-02-25

@author: wf
'''
from tests.basemwtest import BaseMediawikiTest
from meta.smw_type import SMW_Type
from meta.mw import SMWAccess

class TestMetaModel(BaseMediawikiTest):
    """
    test round trip meta model handling
    """
    
    def setUp(self, debug=False, profile=True):
        """
        setUp
        """
        BaseMediawikiTest.setUp(self, debug=debug, profile=profile)
        for wikiId in ["wiki"]:
            self.getSMWAccess(wikiId,save=True)        
            
    def testSMW_Type(self):
        """
        test SMW_Type handling
        """
        debug=True
        smwAccess=self.smwAccessMap["wiki"]
        ask_query=SMW_Type.askQuery()
        did=smwAccess.smw.query(ask_query)
        for record in did.values():
            smw_Type=SMW_Type.fromDict(record)
            if debug:
                print(smw_Type)
            pass
        